import numpy as np
from datetime import datetime, timezone
from keino import Params, k_path
from ristikko import System, Space, models, Symbol, Site

def test_kspace_vs_analytic():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 7)
    params = dict(
            t = rands[0],
            mu = rands[1],
            Jx = rands[2],
            Jy = rands[3],
            Jz = rands[4],
            alpha = rands[5],
            Delta = rands[6],
    )
    params = Params(params)

    def hamiltonian(k):
        t, mu, alpha, J, Delta = params.t, params.mu, params.alpha, params.Jz, params.Delta
        a1, a2 = lattice.lattice_vectors[0], lattice.lattice_vectors[1]
        rashba_matrix = kspace.lattice.rashba_matrix

        def Hk(k):
            ek = 2*t*np.cos(k @ (-a1 + a2)) + 2*t*np.cos(k @ (a1)) + 2*t*np.cos(k @ (a2)) - mu

            ak = [
                + np.exp(1j * k @ a1) * rashba_matrix[0, 0, 1] * alpha,
                - np.exp(-1j * k @ a1) * rashba_matrix[0, 0, 1] * alpha,
                + np.exp(1j * k @ a2) * rashba_matrix[1, 0, 1] * alpha,
                - np.exp(-1j * k @ a2) * rashba_matrix[1, 0, 1] * alpha,
                + np.exp(1j * k @ (-a1 + a2)) * (-rashba_matrix[0, 0, 1] + rashba_matrix[1, 0, 1]) * alpha,
                - np.exp(-1j * k @ (-a1 + a2)) * (-rashba_matrix[0, 0, 1] + rashba_matrix[1, 0, 1]) * alpha,
            ]
            ak = np.array(ak)
            ak = np.sum(ak, axis=0)
            return np.array([
                [ek + J, ak + params.Jx - 1j * params.Jy],
                [np.conj(ak) + params.Jx + 1j * params.Jy, ek - J]
            ])

        Hk1 = Hk(k)
        Hk2 = Hk(-k)
        D = np.array([[0, Delta], [-Delta, 0]])
        H_BdG = np.block([
            [Hk1, D],
            [-D, -Hk2.T]
        ])
        return H_BdG

    k = rng.normal(0, 2*np.pi, 2)

    lattice = models.Triangular()
    kspace = System(Space.KSpace, lattice, params)
    H_BdG = kspace.hamiltonian(k=k)

    ref = hamiltonian(k)

    test = abs(H_BdG - ref).max()

    assert test < 1e-13

def test_kspace_vs_realspace():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 7)
    params = dict(
            N1 = 16,
            N2 = 16,
            t = rands[0],
            mu = rands[1],
            Jx = rands[2],
            Jy = rands[3],
            Jz = rands[4],
            alpha = rands[5],
            Delta = rands[6],
            pbc = True,
    )
    params = Params(params)

    lattice = models.Triangular()
    realspace = System(Space.RealSpace, lattice, params)
    assert realspace == Space.KSpace, f"{seed}, {params}"

def test_kspace_vs_ribbon():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 6)

    params_ribbon = dict(
            N1 = 6,
            N2 = 5,
            L = 5,
            mu = rands[0],
            Jx = rands[1],
            Jy = rands[2],
            Jz = rands[3],
            t = 1,
            alpha = rands[4],
            Delta = rands[5],
            pbc = True,
            direction = "a1",
    )
    params_ribbon = Params(params_ribbon)
    params_kspace = dict(
            N1 = 6,
            N2 = 5,
            mu = rands[0],
            Jx = rands[1],
            Jy = rands[2],
            Jz = rands[3],
            t = 1,
            alpha = rands[4],
            Delta = rands[5],
    )
    params_kspace = Params(params_kspace)

    lattice = models.Triangular()

    syst = System(Space.Ribbon, lattice, params_ribbon)
    kspace = System(Space.KSpace, lattice, params_kspace)

    assert syst == kspace, f"{seed}, {params_ribbon}"

def test_multisite_kspace_vs_kspace():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 6)

    params = dict(
        N1 = 8,
        N2 = 8,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    params = Params(params)
    four_site = models.Triangular(N1=2, N2=2)

    lattice = models.Triangular()
    lattice.add_sublattice(four_site)

    left = System(Space.KSpace, lattice, params)

    params = dict(
        N1 = 16,
        N2 = 16,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    lattice = models.Triangular()
    right = System(Space.KSpace, lattice, params)

    assert left == right, f"{seed}, {params}"

def test_multisite_kspace_vs_realspace1():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 7)

    params = dict(
        N1 = 16,
        N2 = 1,
        t = rands[0],
        mu = rands[1],
        Jx = rands[2],
        Jy = rands[3],
        Jz = rands[4],
        alpha = rands[5],
        Delta = rands[6],
        pbc = True,
    )
    params = Params(params)

    lattice = models.Triangular()
    sublattice = models.Triangular(N1=1, N2=4)
    lattice.add_sublattice(sublattice)

    kspace = System(Space.KSpace, lattice, params)

    lattice = models.Triangular()
    params.N2 = 4
    realspace = System(Space.RealSpace, lattice, params)

    assert kspace == realspace

def test_multisite_kspace_vs_realspace2():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 6)

    params = dict(
        N1 = 8,
        N2 = 8,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    params = Params(params)
    four_site = models.Triangular(N1=2, N2=2)

    lattice = models.Triangular()
    lattice.add_sublattice(four_site)

    left = System(Space.RealSpace, lattice, params)

    params = dict(
        N1 = 16,
        N2 = 16,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    lattice = models.Triangular()
    right = System(Space.KSpace, lattice, params)

    assert left == right, f"{seed}, {params}"

def test_multisite_kspace_vs_realspace3():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 6)

    params = dict(
        N1 = 8,
        N2 = 8,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    params = Params(params)
    four_site = models.Triangular(N1=2, N2=2)

    lattice = models.Triangular()
    lattice.add_sublattice(four_site)

    left = System(Space.KSpace, lattice, params)

    params = dict(
        N1 = 16,
        N2 = 16,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    lattice = models.Triangular()
    right = System(Space.RealSpace, lattice, params)

    assert left == right, f"{seed}, {params}"

def test_multisite_realspace_vs_realspace():
    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 6)

    params = dict(
        N1 = 8,
        N2 = 8,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    params = Params(params)
    four_site = models.Triangular(N1=2, N2=2)

    lattice = models.Triangular()
    lattice.add_sublattice(four_site)

    left = System(Space.RealSpace, lattice, params)

    params = dict(
        N1 = 16,
        N2 = 16,
        mu = rands[0],
        Jx = rands[1],
        Jy = rands[2],
        Jz = rands[3],
        t = 1,
        alpha = rands[4],
        Delta = rands[5],
        pbc = True,
    )
    lattice = models.Triangular()
    right = System(Space.RealSpace, lattice, params)

    assert left == right, f"{seed}, {params}"

def test_impurities_kspace_vs_realspace1():
    # Chain with an impurity attached to each site, also hopping to next site
    # x-x-x-x    impurities
    # |\|\|\|
    # o-o-o-o    chain

    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 9)

    params = dict(
        N1 = 16,
        N2 = 1,
        t = rands[0],
        mu = rands[1],
        Jx = rands[2],
        Jy = rands[3],
        Jz = rands[4],
        alpha = rands[5],
        Delta = rands[6],
        impurities = dict(
            t = rands[7],
            mu = rands[8],
        ),
        pbc = True,
    )
    params = Params(params)

    lattice = models.Triangular()
    sys = System(Space.KSpace, lattice, params)

    t = Symbol("impurities.t")
    mu = Symbol("impurities.mu")
    idx = sys.add_impurity()
    imp = sys.impurities[idx]
    for i in range(2):
        imp.add_hopping(imp, i, i, mu)
        imp.add_hopping(imp, i, i, t, d1=1)
        imp.add_hopping(imp, i, i, t, d1=-1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d1=1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d1=-1)

    realspace = System(Space.RealSpace, lattice, params)
    for x in range(params.N1):
        idx = realspace.add_impurity()
        imp = realspace.impurities[idx]
        for i in range(2):
            imp.add_hopping(imp, i, i, mu)
            if x > 0:
                imp1 = realspace.impurities[idx-1]
                imp.add_hopping(imp1, i, i, t)
            imp.add_hopping(Site(x, 0, 0, 0), i, i, t)
            imp.add_hopping(Site((x+1)%params.N1, 0, 0, 0), i, i, t)
            imp.add_hopping(Site((x-1)%params.N1, 0, 0, 0), i, i, t)

    imp0 = realspace.impurities[0]
    imp1 = realspace.impurities[-1]
    imp0.add_hopping(imp1, 0, 0, t)
    imp0.add_hopping(imp1, 1, 1, t)

    assert sys == realspace, f"{seed}, {params}"

def test_impurities_kspace_vs_realspace2():
    # Impurities centred on a square lattice, coupled to each other and also neighbours
    # o-o-o-o
    # |x|x|x|
    # o-o-o-o
    # |x|x|x|

    seed = int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())
    rng = np.random.default_rng(seed)
    rands = rng.normal(0, 1, 9)

    params = dict(
        N1 = 16,
        N2 = 16,
        t = rands[0],
        mu = rands[1],
        Jx = rands[2],
        Jy = rands[3],
        Jz = rands[4],
        alpha = rands[5],
        Delta = rands[6],
        impurities = dict(
            t = rands[7],
            mu = rands[8],
        ),
        pbc = True,
    )
    params = Params(params)

    lattice = models.Triangular()
    sys = System(Space.KSpace, lattice, params)

    t = Symbol("impurities.t")
    mu = Symbol("impurities.mu")
    idx = sys.add_impurity()
    imp = sys.impurities[idx]
    for i in range(2):
        imp.add_hopping(imp, i, i, mu)
        imp.add_hopping(imp, i, i, t, d1=1)
        imp.add_hopping(imp, i, i, t, d1=-1)
        imp.add_hopping(imp, i, i, t, d2=1)
        imp.add_hopping(imp, i, i, t, d2=-1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d1=1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d2=1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d1=1, d2=1)
        imp.add_hopping(Site(0, 0, 0, 0), i, i, t, d1=-1, d2=-1)

    realspace = System(Space.RealSpace, lattice, params)
    idxs = np.arange(params.N1*params.N2).reshape(params.N1, params.N2)
    for x, y in np.ndindex(params.N1, params.N2):
        idx = realspace.add_impurity()
        imp = realspace.impurities[idx]
        for i in range(2):
            imp.add_hopping(imp, i, i, mu)

            if x > 0:
                imp1 = realspace.impurities[idxs[x-1, y]]
                imp.add_hopping(imp1, i, i, t)
            if y > 0:
                imp1 = realspace.impurities[idxs[x, y-1]]
                imp.add_hopping(imp1, i, i, t)

            imp.add_hopping(Site(x, y, 0, 0), i, i, t)
            imp.add_hopping(Site((x+1)%params.N1, y, 0, 0), i, i, t)
            imp.add_hopping(Site(x, (y+1)%params.N2, 0, 0), i, i, t)
            imp.add_hopping(Site((x+1)%params.N1, (y+1)%params.N1, 0, 0), i, i, t)
            imp.add_hopping(Site((x-1)%params.N1, (y-1)%params.N1, 0, 0), i, i, t)

    for y in range(params.N2):
        imp0 = realspace.impurities[idxs[0, y]]
        imp1 = realspace.impurities[idxs[-1, y]]
        imp0.add_hopping(imp1, 0, 0, t)
        imp0.add_hopping(imp1, 1, 1, t)

    for x in range(params.N1):
        imp0 = realspace.impurities[idxs[x, 0]]
        imp1 = realspace.impurities[idxs[x, -1]]
        imp0.add_hopping(imp1, 0, 0, t)
        imp0.add_hopping(imp1, 1, 1, t)

    assert sys == realspace, f"{seed}, {params}"



# pylint: disable=line-too-long, fixme, too-many-arguments, too-many-positional-arguments, invalid-name, too-many-locals, too-many-instance-attributes, use-dict-literal, bare-except

"""
A System is a representation of a particular Hamiltonian, including
e.g. disorder, impurities, removed sites, particular parameters.

A System is based around a Lattice. A Lattice must be defined before
a System can be created.
"""

import warnings
import numpy as np
from scipy import sparse
from scipy import optimize
from keino import Params, k_path, ndindex
from pfapack import pfaffian as pf
try:
    from pfapack.ctypes import pfaffian as cpf
except:
    from pfapack.pfaffian import pfaffian as cpf
from .lattice import Space, Lattice
from .impurity import Impurity
from .utils import ragged_to_ndarray
from .chern import calculate as calculate_chern
from .geometry import Geometry
from .models import Square, Triangular, Graphene

def _realspace(N1, N2, n_orbitals, n_impurities, ts, ij):
    tot = ts.size
    _ij = np.zeros((2, tot), np.int64)
    v = np.zeros((tot), np.complex128)

    ind1 = np.arange(N1 * N2 * n_orbitals).reshape(N1, N2, n_orbitals)
    ind2 = np.arange(N1 * N2 * n_orbitals, N1 * N2 * n_orbitals + N1 * N2 * n_orbitals * n_impurities).reshape(N1, N2, n_orbitals*n_impurities)
    ind = np.concat((ind1, ind2), axis=2)

    for n in range(ts.size):
        x0, y0, x1, y1, i, j = ij[n]
        v[n] = ts[n]
        _ij[0, n] = ind[x0, y0, i]
        _ij[1, n] = ind[x1, y1, j]

    return v, _ij

def _kspace(n_orbitals, ij, ts):
    H0 = np.zeros((n_orbitals, n_orbitals), np.complex128)
    for n in range(ts.size):
        i, j = ij[n]
        t = ts[n]

        H0[i, j] += t
    return H0

def _ribbon(L, n_orbitals, n_impurities, ts, ij):
    tot = len(ts)
    _ij = np.zeros((2, tot), np.int64)
    v = np.zeros((tot), np.complex128)

    ind1 = np.arange(L * n_orbitals).reshape(L, n_orbitals)
    ind2 = np.arange(L * n_orbitals, L * n_orbitals + L * n_orbitals * n_impurities).reshape(L, n_orbitals*n_impurities)
    ind = np.concat((ind1, ind2), axis=1)

    for n in range(ts.size):
        x0, x1, i, j = ij[n]
        v[n] = ts[n]
        _ij[0, n] = ind[x0, i]
        _ij[1, n] = ind[x1, j]

    return v, _ij

class System():
    def __init__(self, space: Space, lattice: Lattice, params: Params, spinful=True, n_orbitals=None):
        # pylint: disable=no-member, too-many-branches
        warnings.simplefilter("ignore", category=np.exceptions.ComplexWarning)

        if isinstance(lattice, str):
            match lattice:
                case "square":
                    lattice = Square()
                case "triangular":
                    lattice = Triangular()
                case "graphene":
                    lattice = Graphene()
                case _:
                    raise ValueError(f"Unknown lattice {lattice}!")

        self.lattice = lattice.copy()
        if self.lattice.n_orbitals is not None:
            n_orbitals = self.lattice.n_orbitals

        if n_orbitals is None:
            self.spinful = spinful
            self.n_orbitals = 2 if spinful else 1
        else:
            self.spinful = not bool(n_orbitals)
            self.n_orbitals = n_orbitals

        self.params = Params(params)
        self.space = Space(space)
        self.lattice.set_n_orbitals(self.n_orbitals)
        self.lattice_vectors = self.lattice.lattice_vectors
        self.recip_vectors = 2 * np.pi * np.eye(2) @ np.linalg.inv(self.lattice_vectors.T)

        self.N1 = -1
        self.N2 = -1
        self.L = -1
        match self.space:
            case Space.RealSpace:
                self.N1 = self.params.N1
                self.N2 = self.params.N2
            case Space.Ribbon:
                self.L = self.params.L
                if "N1" in self.params :
                    self.N1 = self.params.N1
                if "N2" in self.params:
                    self.N2 = self.params.N2
            case Space.KSpace:
                self.N1 = 1
                self.N2 = 1
                if "N1" in self.params :
                    self.N1 = self.params.N1
                if "N2" in self.params:
                    self.N2 = self.params.N2
            case _:
                raise ValueError(f"Space {self.space} not implemented")

        self.lattice.N1 = self.N1
        self.lattice.N2 = self.N2
        if space == Space.Ribbon:
            self.lattice.N1 = self.lattice.N2 = self.lattice.L = self.L
        if space == Space.KSpace:
            self.lattice.N1 = self.lattice.N2 = 1

        if self.is_superconducting():
            if np.isscalar(self.params.Delta):
                # Assume s-wave superconductivity
                self.params.Delta = Params(dict(s={'0': self.params.Delta, '1': 0, '2': 0}, p={'0': 0, '1': 0}, d={'0': 0, '1': 0}, f={'0': 0, '1': 0}))

        self.impurities = []
        self.geometry = None

    def shape(self):
        return self.lattice.shape()

    def get_block_dimension(self):
        """Return the size of each block in the Hamiltonian"""
        _, _, N1_sub, N2_sub = self.lattice.shape()
        dim = N1_sub * N2_sub * self.n_orbitals
        return dim

    def get_dimension(self):
        """Return the overall dimension of the Hamiltonian matrix"""
        match self.space:
            case Space.KSpace:
                return self.get_block_dimension() + self.n_orbitals * len(self.impurities)
            case Space.Ribbon:
                return self.get_block_dimension() * self.L + self.n_orbitals * len(self.impurities)
            case Space.RealSpace:
                return self.get_block_dimension() * self.N1 * self.N2 + self.n_orbitals * len(self.impurities)
            case _:
                raise ValueError(f"Space {self.space} not implemented")

    def get_dimension_split(self):
        """
        Return the overall dimension of the Hamiltonian matrix, split
        into the dimension of the underlying lattice, and the dimension
        of any impurities
        """
        match self.space:
            case Space.KSpace:
                return self.get_block_dimension(), self.n_orbitals * len(self.impurities)
            case Space.Ribbon:
                return self.get_block_dimension() * self.L, self.n_orbitals * len(self.impurities)
            case Space.RealSpace:
                return self.get_block_dimension() * self.N1 * self.N2, self.n_orbitals * len(self.impurities)
            case _:
                raise ValueError(f"Space {self.space} not implemented")

    def coords(self):
        _ = self.lattice(self.params, self.space)
        return self.lattice.coords

    def is_superconducting(self):
        return bool(self.params.get("Delta", False))

    def add_impurity(self, coord=None):
        """
        Add an impurity to the system.

        Parameters
        ----------
        coord: pair[float]
            Optional, coordinate of the impurity. Used for plotting

        Returns
        ----------
        The unique index of the impurity
        """
        idx = len(self.impurities)
        N1, N2, N1_sub, N2_sub = self.shape()
        impurity = Impurity(idx, N1, N2, N1_sub, N2_sub, spinful=self.spinful, coord=coord)
        self.impurities.append(impurity)
        return idx

    def set_geometry(self, shape, sites=None, coords=None):
        N1, N2, N1_sub, N2_sub = self.shape()
        self.geometry = Geometry(N1, N2, N1_sub, N2_sub, shape, self.lattice.bravais_lattice, self.params.impurities, self.space, self.spinful, sites=sites, coords=coords)
        self.impurities = self.geometry()

    def hamiltonian(self, *args, **kwargs):
        # pylint: disable=inconsistent-return-statements
        """Construct the Hamiltonian matrix"""

        match self.space:
            case Space.RealSpace:
                return self.hamiltonian_realspace(*args, **kwargs)
            case Space.KSpace:
                return self.hamiltonian_kspace(*args, **kwargs)
            case Space.Ribbon:
                return self.hamiltonian_ribbon(*args, **kwargs)
            case _:
                raise ValueError(f"Space {self.space} not implemented")

    def hamiltonian_realspace(self, dense=False, test_hermitian=False, postprocess_hoppings=None, trim=False):
        ts, ij = self.lattice(self.params, self.space, impurities=self.impurities, trim=trim)

        n_orbitals = self.get_block_dimension()
        if postprocess_hoppings is not None:
            ts, ij = postprocess_hoppings(self.params, ts, ij)

        N1, N2, _, _ = self.shape()
        dim = self.get_dimension()

        v, ij = _realspace(N1, N2, n_orbitals, len(self.impurities), ts, ij)
        H = sparse.coo_matrix((v, ij), shape=(dim, dim))

        if test_hermitian:
            test = np.abs(H - H.conj().T).max()
            assert test < 1e-12, test

        if not self.is_superconducting():
            if dense:
                H = H.toarray()
                if trim:
                    H = H[~np.all(H == 0, axis=1)]
                    H = H[:, ~np.all(H == 0, axis=0)]
            return H

        sc_ts, sc_ij = self.lattice(self.params, self.space, superconducting=True)
        v, ij = _realspace(N1, N2, n_orbitals, len(self.impurities), sc_ts, sc_ij)
        D = sparse.coo_matrix((v, ij), shape=(dim, dim))

        H_BdG = sparse.bmat([
            [H, D],
            [-D, -H.T]
        ])
        H_BdG.eliminate_zeros()

        if dense:
            if trim:
                H_BdG = H_BdG[~np.all(H_BdG == 0, axis=1)]
                H_BdG = H_BdG[:, ~np.all(H_BdG == 0, axis=0)]
            return H_BdG.toarray()

        return H_BdG

    def hamiltonian_kspace(self, k, test_hermitian=False, trim=False, postprocess_hoppings=None, **kwargs):
        # pylint: disable=unused-argument
        # dense argument is needed for compatibility
        def block(k):
            ts, ij = self.lattice(self.params, self.space, k=k, direction=direction, impurities=self.impurities)
            if postprocess_hoppings is not None:
                ts, ij = postprocess_hoppings(self.params, ts, ij)

            Hk = _kspace(dim, ij, ts)
            return Hk

        direction = self.params.get("direction", "both")
        dim = self.get_dimension()
        k = np.array(k)

        Hk1 = block(k)

        if test_hermitian:
            test = np.abs(Hk1 - Hk1.conj().T).max()
            assert test < 1e-12, test

        if not self.is_superconducting():
            # Remove empty rows and cols
            if trim:
                Hk1 = Hk1[~np.all(Hk1 == 0, axis=1)]
                Hk1 = Hk1[:, ~np.all(Hk1 == 0, axis=0)]
            return Hk1

        sc_ts, sc_ij = self.lattice(self.params, self.space, k=k, superconducting=True, direction=direction)
        D = _kspace(dim, sc_ij, sc_ts)
        tmp = ragged_to_ndarray([Hk1, D])
        Hk1, D = tmp[0], tmp[1]

        Hk2 = block(-k)
        H_BdG = np.block([
            [Hk1, D],
            [-D, -Hk2.T]
        ])

        # Remove empty rows and cols
        if trim:
            H_BdG = H_BdG[~np.all(H_BdG == 0, axis=1)]
            H_BdG = H_BdG[:, ~np.all(H_BdG == 0, axis=0)]
        return H_BdG

    def hamiltonian_ribbon(self, k, test_hermitian=False, dense=False, postprocess_hoppings=None):
        def block(k):
            ts, ij = self.lattice(self.params, self.space, k=k, direction=direction, impurities=self.impurities)

            if postprocess_hoppings is not None:
                ts, ij = postprocess_hoppings(self.params, ts, ij)

            v, ij = _ribbon(self.L, n_orbitals, len(self.impurities), ts, ij)
            Hk = sparse.coo_matrix((v, ij), shape=(dim, dim))
            Hk.eliminate_zeros()
            return Hk

        direction = self.params.get("direction", "a1")
        n_orbitals = self.get_block_dimension()
        dim = self.get_dimension()

        Hk1 = block(k)

        if test_hermitian:
            test = np.abs(Hk1 - Hk1.conj().T).max()
            assert test < 1e-12, test

        if not self.is_superconducting():
            if dense:
                return Hk1.toarray()
            return Hk1

        Hk2 = block(-k)

        ts, ij = self.lattice(self.params, self.space, superconducting=True, k=k, direction=direction)
        v, ij = _ribbon(self.L, n_orbitals, len(self.impurities), ts, ij)
        D = sparse.coo_matrix((v, ij), shape=(dim, dim))
        H_BdG = sparse.bmat([
            [Hk1, D],
            [-D, -Hk2.T]
        ])

        if dense:
            H_BdG = H_BdG.toarray()

        return H_BdG

    def plot(self, axis=None, colours=None, axis_scale=5, site_scale=3, aspect=None, filename=None):
        # pylint: disable=import-outside-toplevel, raise-missing-from, too-many-branches, too-many-statements, unused-variable, import-error, too-many-function-args, no-member
        """Plot the system"""
        try:
            import matplotlib.pyplot as plt
            from matplotlib.collections import LineCollection
        except:
            raise ValueError("matplotlib is not installed!")

        impurities = self.impurities if len(self.impurities) > 0 else None
        ts, ij = self.lattice(self.params, Space.RealSpace, impurities=impurities)
        hd = self.lattice.hd
        dim = self.get_block_dimension()

        N1_super, N2_super, N1_sub, N2_sub = self.shape()

        _aspect = N2_super / N1_super * self.lattice.aspect
        aspect = _aspect if aspect is None else aspect

        if hasattr(self.lattice.plot_scale, "__len__"):
            a1 = self.lattice.plot_scale[0] * self.lattice.lattice_vectors[0]
            a2 = self.lattice.plot_scale[1] * self.lattice.lattice_vectors[1]
        else:
            a1, a2 = self.lattice.plot_scale * self.lattice.lattice_vectors

        if axis is None:
            fig, axis = plt.subplots(figsize=(axis_scale, axis_scale*aspect))

        if colours is None:
            colours = "black"
        else:
            _colours = np.array(colours).reshape(N1_sub, N2_sub)
            colours = np.empty((N1_super, N1_sub, N2_super, N2_sub), object)
            for i, j in ndindex(N1_sub, N2_sub):
                colours[:, i, :, j] = _colours[i, j]
            colours = colours.reshape(N1_super*N1_sub*N2_super*N2_sub)
            colours = [str(c) for c in colours]

        scale = site_scale * np.ones(N1_super*N2_super*N1_sub*N2_sub).reshape(N1_super, N2_super, N1_sub, N2_sub)
        for site in self.lattice.removed_sites:
            x, y, x_sub, y_sub = site
            scale[x, y, x_sub, y_sub] = 0
        scale = scale.reshape(N1_super*N2_super*N1_sub*N2_sub)

        axis.scatter(self.lattice.coords[:, 0], self.lattice.coords[:, 1], scale, colours)

        if self.impurities is not None:
            for impurity in self.impurities:
                x, y = impurity.coord[0], impurity.coord[1]
                x, y = x * a1 + y * a2
                axis.scatter(x, y, 30, "tab:red", marker="*")

        def coords_in(x, y, eps=1e-12):
            for i, j in self.lattice.coords:
                if (i-x) <= eps and (j-y) <= eps:
                    return True
            return False

        def impurity_coords(i):
            # idx = (i % dim) // self.n_orbitals
            if self.spinful:
                if i % 2:
                    idx = (i - dim - 1) // 2
                else:
                    idx = (i - dim) // 2
            else:
                idx = i - dim
            idx = int(idx)
            return self.impurities[idx].coord

        # slv = self.lattice.sublattice.lattice_vectors
        lines = []
        lines_imp = []
        lws = []
        lws_imp = []
        for n, t in enumerate(ts):
            lw = 0.5 * t
            x0, y0, x1, y1, i, j = ij[n]

            if i >= dim:
                x0, y0 = impurity_coords(i)
            if j >= dim:
                x1, y1 = impurity_coords(j)

            if self.lattice.basis is not None:
                if i < dim:
                    basis0 = self.lattice.basis[i]
                    x0, y0 = x0 * N1_sub * a1 + y0 * N2_sub * a2
                    x0, y0 = (x0, y0) + basis0
                    # x0, y0 = (x0, y0) + slv @ basis0
                else:
                    x0, y0 = x0 * a1 + y0 * a2
                if j < dim:
                    basis1 = self.lattice.basis[j]
                    x1, y1 = x1 * N1_sub * a1 + y1 * N2_sub * a2
                    x1, y1 = (x1, y1) + basis1
                    # x1, y1 = (x1, y1) + slv @ basis1
                else:
                    x1, y1 = x1 * a1 + y1 * a2
            else:
                x0, y0 = x0 * a1 + y0 * a2
                x1, y1 = x1 * a1 + y1 * a2

            if not self.params.pbc and not coords_in(x1, y1):
                lw = 0
            if not self.params.pbc and not coords_in(x0, y0):
                lw = 0
            if i < dim and j < dim:
                if abs(lw) > 0:
                    lines.append([(x0, y0), (x1, y1)])
                    lws.append(lw)
            else:
                if abs(lw) > 0:
                    lines_imp.append([(x0, y0), (x1, y1)])
                    lws_imp.append(lw)

        lc = LineCollection(lines, colors="black", linewidths=lws)
        axis.add_collection(lc)

        lc = LineCollection(lines_imp, colors="red", linewidths=lws_imp)
        axis.add_collection(lc)

        if filename:
            plt.savefig(filename, bbox_inches="tight")
        plt.show()

    def __eq__(self, other, threshold=1e-13, debug=False):
        """Check if two Systems are equivalent. Useful for testing."""
        if isinstance(other, str):
            other = Space(other)

        if isinstance(other, Space):
            other = System(other, self.lattice, self.params)
            if self.geometry is not None or len(self.impurities) > 0:
                warnings.warn("Warning: geometry or impurities set. self and other may not have consistent shapes.")
        elif other.space == self.space:
            dim0 = self.get_block_dimension() * self.N1 * self.N2 + self.n_orbitals * len(self.impurities)
            dim1 = other.get_block_dimension() * other.N1 * other.N2 + other.n_orbitals * len(other.impurities)
            assert dim0 == dim1, f"{dim0} != {dim1}"

        def calc_spectra(sys):
            if sys.space == Space.RealSpace:
                E = sys.diagonalise(dense=True, cache=False)

            if sys.space == Space.KSpace:
                sym_points1 = np.array([
                    [0, 0],
                    [1, 0],
                ])
                sym_points2 = np.array([
                    [0, 0],
                    [0, 1],
                ])

                points1 = np.dot(sym_points1, sys.recip_vectors)
                points2 = np.dot(sym_points2, sys.recip_vectors)
                dim = sys.get_dimension()
                if sys.is_superconducting():
                    dim *= 2

                X = np.array(k_path(sys.N1, points1, endpoint=False))
                Y = np.array(k_path(sys.N2, points2, endpoint=False))

                E = np.zeros((sys.N1, sys.N2, dim))
                for x, y in np.ndindex(sys.N1, sys.N2):
                    k = X[x] + Y[y]
                    Hk = sys.hamiltonian(k, test_hermitian=True)
                    if Hk.shape == (0, 0):
                        continue
                    E[x, y] = np.linalg.eigvalsh(Hk)
                E = np.sort(E.flatten())

            if sys.space == Space.Ribbon:
                sym_points1 = np.array([
                    [0, 0],
                    [1, 0],
                ])
                sym_points2 = np.array([
                    [0, 0],
                    [0, 1],
                ])

                points1 = np.dot(sym_points1, sys.recip_vectors)
                points2 = np.dot(sym_points2, sys.recip_vectors)

                X = np.array(k_path(sys.N1, points1, endpoint=False))
                Y = np.array(k_path(sys.N2, points2, endpoint=False))

                dim = sys.get_dimension()
                if self.is_superconducting():
                    dim *= 2

                nk = sys.N1 if sys.params.direction == "a1" else sys.N2
                E = np.zeros((nk, dim))
                for n in range(nk):
                    k = X[n] if sys.params.direction == "a1" else Y[n]
                    H = sys.hamiltonian(k, dense=True)
                    E[n] = np.linalg.eigvalsh(H)
                E = np.sort(E.flatten())

            return E

        if not self.params.get("pbc", True):
            warnings.warn("Warning: open boundaries detected")

        E = calc_spectra(self)
        E_other = calc_spectra(other)
        test = abs(E - E_other).max()

        if debug:
            print(test)
            return E, E_other

        test = test < threshold
        return test

    def diagonalise(self, return_eigenvectors=False, dense=False, n_energies=20, k=None, cache=True, refresh_cache=False, retry=False, always_succeed=False, sigma=0, **kwargs):
        # pylint: disable=too-many-branches, too-many-statements, import-outside-toplevel, unnecessary-lambda-assignment
        """
        Diagonalise the Hamiltonian matrix.
        This function optionally can cache the result using joblib.

        Parameters
        ----------
        return_eigenvectors: bool
        dense: bool
            Whether to construct then diagonalise the Hamiltonian matrix
            as a dense matrix (all eigenvalues) or sparse (n_energies around E = sigma)
        n_energies: int
            Number of energies to return, if doing a sparse diagonalisation
        sigma: float
            energy around which to diagonalise (sparse)
        k: float | pair[float] | array[float] | array[pair[float]]
            momentum to diagonalise around (only if self.space == Space.KSpace)
        cache: bool
            Cache the results
        refresh_cache: bool
            Ignore cached result and recalculate
        retry:
            Sometimes a sparse diagonalisation will fail. If it does and
            retry == True, try to do a dense diagonalisation. Useful for
            unsupervised (i.e. on a cluster) jobs.
        always_succeed: bool
            If the diagonalisation fails, throw away the error and return
            nan arrays with the correct shape. Useful for unsupervised
            (i.e. on a cluster) jobs
        kwargs:
            passed to self.hamiltonian()

        Returns
        ----------
        Either the eigenvalues or the eigenvalues and eigenvectors
        """
        if cache:
            from joblib import Memory
            memory = Memory("cache", verbose=0)
            if refresh_cache:
                cache_validation_callback = lambda x: False
            else:
                cache_validation_callback = lambda x: True
            cache = memory.cache(cache_validation_callback=cache_validation_callback)
        else:
            cache = lambda f: f

        @cache
        def work(params, lattice, space, dense, k, dim):
            # pylint: disable= unused-argument, broad-exception-caught
            # Here need arguments lattice, space, params to cache correctly
            if self.space == Space.RealSpace:
                H = self.hamiltonian(dense=dense, **kwargs)
            else:
                H = self.hamiltonian(dense=dense, k=k, **kwargs)

            if dense:
                try:
                    en, wf = np.linalg.eigh(H)
                except Exception as e:
                    if always_succeed:
                        en = np.empty(dim)
                        wf = np.empty((dim, dim), complex)
                    else:
                        raise ValueError(f"Could not diagonalise dense matrix: {e}") from e
            else:
                try:
                    en, wf = sparse.linalg.eigsh(H, sigma=sigma, k=n_energies)
                except Exception as e:
                    if retry:
                        return work(self.params, self.lattice, self.space, dense=True, k=k, dim=dim)
                    if always_succeed:
                        en = np.empty(dim)
                        wf = np.empty((dim, dim), complex)
                    else:
                        raise ValueError(f"Could not diagonalise sparse matrix: {e}") from e
                idx = np.argsort(en)
                en = en[idx]
                wf = wf[:, idx]
            return en, wf

        if self.space == Space.KSpace:
            dense = True

        dim0 = self.get_dimension()
        dim1 = n_energies
        if not dense:
            dim = dim0 if dim0 < dim1 else dim1
        else:
            dim = dim0

        if self.is_superconducting():
            dim *= 2

        match self.space:
            case Space.KSpace | Space.Ribbon:
                if hasattr(k, "__len__"):
                    k = np.squeeze(k)
                    assert len(k.shape) in [1, 2]
                    if len(k.shape) == 1 and k.shape[0] == 2:
                        k = k[None, :]
                    if len(k.shape) == 1:
                        nk = k.size
                    else:
                        nk, _ = k.shape

                    en = np.zeros((nk, dim))
                    wf = np.zeros((nk, dim, dim), complex)
                    for i in range(nk):
                        _en, _wf = work(self.params, self.lattice, self.space,dense, k[i], dim)
                        if len(_en) == dim:
                            en[i], wf[i] = _en, _wf
                else:
                    en, wf = work(self.params, self.lattice, self.space,dense, k, dim)
            case Space.RealSpace:
                en, wf = work(self.params, self.lattice, self.space,dense, k, dim)
            case _:
                raise ValueError(f"Space {self.space} not implemented")

        if return_eigenvectors:
            return en, wf

        return en

    def calc_ldos(self, omega, delta=1e-3, **kwargs):
        """
        Calculate the Local Density of States

        Parameters
        ----------
        omega: float | numpy.ndarray
            energies to calculate LDOS for
        delta: float
            Lorentzian broadening to apply
        kwargs:
            passed to self.diagonalise()

        Returns
        ----------
        The LDOS, reshaped to match the basis
        """
        if self.space is not Space.RealSpace:
            raise ValueError("LDOS can only be calculated in Real Space!")

        en, wf = self.diagonalise(return_eigenvectors=True, dense=True, **kwargs)

        abs2wf = np.abs(wf)**2
        kernel = delta / (np.pi * ((en[:, None] - omega[None, :])**2 + delta**2))
        ldos = abs2wf.dot(kernel)
        ldos = np.squeeze(ldos)
        dim = self.get_dimension()

        if self.is_superconducting():
            ldos = ldos.reshape(2, dim, len(omega))
        else:
            ldos = ldos.reshape(dim, len(omega))

        n_impurities = len(self.impurities)
        N1_super, N2_super, N1_sub, N2_sub = self.shape()
        if n_impurities > 0:
            dim0, _ = self.get_dimension_split()
            if self.is_superconducting():
                ldos0 = ldos[:, :dim0].reshape(2, N1_super, N2_super, N1_sub, N2_sub, self.n_orbitals, len(omega))
                ldos1 = ldos[:, dim0:].reshape(2, len(self.impurities), self.n_orbitals, len(omega))
            else:
                ldos0 = ldos[:dim0].reshape(2, N1_super, N2_super, N1_sub, N2_sub, self.n_orbitals, len(omega))
                ldos1 = ldos[dim0:].reshape(len(self.impurities), self.n_orbitals, len(omega))
            ldos = (ldos0, ldos1)
        else:
            if self.is_superconducting():
                ldos = ldos.reshape(2, N1_super, N2_super, N1_sub, N2_sub, self.n_orbitals, len(omega))
            else:
                ldos = ldos.reshape(N1_super, N2_super, N1_sub, N2_sub, self.n_orbitals, len(omega))

        return ldos

    def calc_chern(self, Nk=21, offset=0, n_bands=None, **kwargs):
        """Calculate the Chern number with the Fukui-Hatsugai method"""
        assert self.space == Space.KSpace

        Kx = np.linspace(-np.pi+offset, np.pi+offset, Nk)
        Ky = np.linspace(-np.pi+offset, np.pi+offset, Nk)

        if n_bands is None:
            n_bands = self.get_dimension()

        bands = [{} for n in range(n_bands)]

        for x_index, x in enumerate(Kx):
            for y_index, y in enumerate(Ky):
                k = [(x, y)]
                _, wf = self.diagonalise(return_eigenvectors=True, k=k, **kwargs)
                for i in range(n_bands):
                    bands[i][(x_index, y_index)] = wf[0, :, i]

        Cs = [calculate_chern(band, Nk) for band in bands]
        C = sum(Cs)
        C = np.round(C)
        return C

    def calc_majorana(self, do_tests=False, fast=True):
        """Calculate the Majorana number"""
        dim = self.get_dimension()

        U = np.block([
            [np.eye(dim), np.eye(dim)],
            [-1j * np.eye(dim), 1j * np.eye(dim)]
        ])

        sym_points = [[0, 0], [0.5, 0]]
        points = np.dot(sym_points, self.recip_vectors)

        H = [
            self.hamiltonian(points[0]),
            self.hamiltonian(points[1])
        ]

        A = [
            -0.5j * U @ H[0] @ U.T.conj(),
            -0.5j * U @ H[1] @ U.T.conj(),
        ]

        if do_tests:
            test = abs((A[0] + A[0].T).max())
            assert test < 4e-14, f'{test}'
            test = abs((A[1] + A[1].T).max())
            assert test < 4e-14, f'{test}'

        if fast:
            Pf = [
                cpf(A[0]),
                cpf(A[1])
            ]
        else:
            Pf = [
                pf.pfaffian(A[0]),
                pf.pfaffian(A[1])
            ]

        if do_tests:
            left = Pf[0]**2
            right = np.linalg.det(A[0])
            assert np.isclose(left, right), f'{Pf[0]}, {left}, {right}'
            left = Pf[1]**2
            right = np.linalg.det(A[1])
            assert np.isclose(left, right), f'{Pf[1]}, {left}, {right}'

        M = np.sign(Pf[0] * Pf[1]).real
        return M

    def winding_number(self, do_zk=False, do_theta_k=True, Nk=128):
        """
        Calculate the winding number,

        Parameters
        ----------
        do_zk, do_theta_k: bool
            Specify method to use
        Nk: int
            number of momenta-space points

        Returns
        ----------
        The winding number
        """
        def Vk(k):
            U = np.block([
                [np.eye(dim), np.eye(dim)],
                [-1j * np.eye(dim), 1j * np.eye(dim)]
                ])
            H = self.hamiltonian(k=k)
            Ak = -0.5j * U @ H @ U.conj().T
            Vk = Ak[:dim, dim:]
            return Vk

        def zk(k):
            detVk = np.linalg.det(Vk(k))
            zk = detVk / np.abs(detVk)
            return zk

        def theta_k(k):
            return -1j * np.log(np.linalg.det(Vk(k)))

        def shift_angle(theta):
            m = np.floor(np.diff(theta, prepend=theta[-1]) / (2*np.pi) + 0.5)
            return m.sum()

        assert self.is_superconducting()
        assert self.space == Space.KSpace
        assert do_zk or do_theta_k
        dim = self.get_dimension()

        direction = self.params.get("direction", "a1")
        if direction == "a1":
            sym_points1 = np.array([
                [0, 0],
                [1, 0],
            ])
        if direction == "a2":
            sym_points2 = np.array([
                [0, 0],
                [0, 1],
            ])

        points = np.dot(sym_points, self.recip_vectors)
        Ks = k_path(Nk, points)

        if do_zk:
            intg = [zk(k) for k in Ks]
            W = shift_angle(np.angle(intg))

        if do_theta_k:
            intg = [theta_k(k) for k in Ks]
            W = shift_angle(np.real(intg))

        return W

    def calc_gap(self, Ks=None, **kwargs):
        """Calculate the gap of the spectra"""
        match self.space:
            case Space.KSpace:
                if Ks is not None:
                    Ek = self.diagonalise(k=Ks, **kwargs)
                    gap = abs(Ek.flatten()).min()
                else:
                    dim = self.get_dimension()
                    x0 = (0, 0)

                    def fun(k):
                        E = self.diagonalise(k=k, **kwargs)
                        return E[0][dim]

                    res = optimize.minimize(fun, x0, method="COBYLA")
                    gap = res.fun
            case Space.RealSpace:
                en = self.diagonalise(**kwargs)
                gap = abs(en).min()
            case _:
                raise ValueError(f"Space {self.space} is not supported")

        return gap

    def superconducting_correlations(self, **kwargs):
        """
        Calculate superconducting correlations
        """
        assert self.is_superconducting()

        match self.space:
            case Space.KSpace:
                X = 2 * np.pi / self.params.N1 * np.arange(N1)
                Y = 2 * np.pi / self.params.N2 * np.arange(N2)
                dim = self.get_dimension()
                Ek = np.empty((self.params.N1, self.params.N2, 2*dim))
                nk = np.empty((self.params.N1, self.params.N2, 2*dim, 2*dim), complex)
                for i, j in np.ndindex(self.params.N1, self.params.N2):
                    x = X[i]
                    y = Y[j]
                    k = (x, y)
                    Ek[i, j], nk[i, j] = self.diagonalise(return_eigenvectors=True, **kwargs)

                Pk = np.empty((self.params.N1, self.params.N2, 2*dim, 2*dim), complex)
                for i, j in np.ndindex(self.params.N1, self.params.N2):
                    Pk[i, j] = nk[i, j, :, :dim].dot(nk[i, j, :, :dim].conj().T)

                Pk = Pk.reshape(self.params.N1, self.params.N2, 2, dim, 2, 2, dim, 2)
                return Pk

            case Space.RealSpace:
                dim = self.get_dimension()
                en, wf = self.diagonalise(return_eigenvectors=True, **kwargs)
                Pk = wf[:, :dim].dot(wf[:, :dim].conj().T)
                return Pk
            case _:
                raise ValueError(f"Space {self.space} not implemented!")

    def fermi_velocities(self):
        def get_fermi_energy_crossings(en):
            crossings = []
            for i, (previous, current) in enumerate(zip(en, en[1:])):
                if previous > 0 and current < 0:
                    crossings.append((i, i+1))
                elif previous < 0 and current > 0:
                    crossings.append((i, i+1))
            return crossings

        X = np.linspace(-np.pi, np.pi, self.params.N1)
        Y = np.linspace(-np.pi, np.pi, self.params.N2)
        Ks = np.meshgrid(X, Y)

        p = self.params.copy()
        self.params.pop("Delta")

        Ek = self.diagonalise(k=Ks)
        velocities = []
        for en in Ek.T:
            crossings = get_fermi_energy_crossings(en)
            crossings = [c for c in crossings if len(c) > 0]

            for i, j in crossings:
                Es = en[i:j+1]
                vf = (Es[1] - Es[0]) / (Ks[1, 1] - Ks[0, 1])
                velocities.append(vf)
        return np.array(velocities)

    def coherence_length(self):
        X = np.linspace(-np.pi, np.pi, self.params.N1)
        Y = np.linspace(-np.pi, np.pi, self.params.N2)
        Ks = np.meshgrid(X, Y)

        gap = self.calc_gap(Ks)
        vF = self.fermi_velocities()

        coherence_lengths = vF / gap
        coherence_lengths = np.round(coherence_lengths, 10)
        coherence_lengths = np.array(list(set(abs(coherence_lengths))))
        return min(coherence_lengths)



# pylint: disable=missing-module-docstring, invalid-name, line-too-long, too-many-instance-attributes, too-many-arguments, too-many-positional-arguments, missing-class-docstring
from dataclasses import dataclass
import numpy as np
from keino import Params
from .lattice import Space, Symbol, Site

class Impurity():
    """
    Represents an impurity in a Hamiltonian.

    An impurity is defined by an energy, and which sites it hops to.
    It also has a nominal position, but this is only used for plotting.
    Impurities are included in the Hamiltonian matrix as additional
    rows and columns.
    """

    def __init__(self, idx, N1, N2, N1_sub, N2_sub, spinful=True, coord=None):
        """
        Parameters
        ----------

        idx: int
            unique index of the impurity. Automatically assigned by
            System.add_impurity()
        N1, N2, N1_sub, N2_sub: int
            shape of the associated system
        spinful: bool
            controls the number of orbitals of the impurity
        coord: pair[float]
            Optional, the position of the impurity. Used for plotting.
        """
        self.idx = idx
        self.n_orbitals = 2 if spinful else 1
        self.n_orbitals_substrate = self.n_orbitals * N1_sub * N2_sub
        self.hoppings = []
        self.N1 = N1
        self.N2 = N2
        self.N1_sub = N1_sub
        self.N2_sub = N2_sub
        self.x = 0
        self.y = 0
        self.coord = coord

    def add_hopping(self, other, i, j, t, d1=0, d2=0):
        """
        Add a hopping from the impurity.

        Parameters:
            other: Impurity | size-4 list/tuple/ndarray
            Where the hopping is going to
                - either another impurity, or
                - a coordinate on the substrate + sublattice (x, y, x_sub, y_sub)
        i, j: int
            The orbitals involved in the hopping
            i = 0: hopping from up orbital
            i = 1: hopping from down orbital
            j = 0: hopping to up orbital
            j = 1: hopping to down orbital
        t: Symbol
            hopping energy
        d1, d2: int
            (For kspace)
            The number of unit cells in the a1 and a2 directions
            that the hopping traverses
        """
        assert isinstance(other, (Impurity, Site))

        if isinstance(other, Impurity):
            j = other.n_orbitals_substrate + other.n_orbitals * other.idx + j

        ind = np.arange(self.n_orbitals_substrate).reshape(self.N1_sub, self.N2_sub, self.n_orbitals)
        if isinstance(other, Site):
            j = ind[other.x_sub, other.y_sub, j]

        i = self.n_orbitals_substrate + self.n_orbitals * self.idx + i
        self.hoppings.append(Hopping(other, i, j, t, d1, d2))

    def __call__(self, params: Params, space, k=None, direction=None, lattice_vectors=None):
        match space:
            case Space.RealSpace:
                ts = [hop.t.replace(params, None, None) for hop in self.hoppings]
                return ts
            case Space.KSpace:
                ts = [hop.t.replace(params, None, None) * np.exp(1j * k @ (hop.d1 * lattice_vectors[0] + hop.d2 * lattice_vectors[1])) for hop in self.hoppings]
                return ts
            case Space.Ribbon:
                ts = []
                if direction == "a1":
                    for hop in self.hoppings:
                        if hop.d2 == 0:
                            t = hop.t.replace(params, None, None) * np.exp(1j * k @ (hop.d1 * lattice_vectors[0]))
                            ts.append(t)
                if direction == "a2":
                    for hop in self.hoppings:
                        if hop.d1 == 0:
                            t = hop.t.replace(params, None, None) * np.exp(1j * k @ (hop.d2 * lattice_vectors[1]))
                            ts.append(t)
                return ts
            case _:
                raise ValueError(f"Space {space} not implemented")

@dataclass
class Hopping:
    other: Site | Impurity
    i: int
    j: int
    t: Symbol
    d1: int
    d2: int



# pylint: disable=line-too-long, too-many-instance-attributes

import numpy as np
from ..lattice import Lattice, BravaisLattice, Symbol
from .square import Square

class Graphene(Lattice):
    def __init__(self, spinful=False, **kwargs):
        super().__init__(BravaisLattice.Triangular, **kwargs)
        self.n_orbitals = 2 if spinful else 1

        sublattice = Square(N1=1, N2=2)
        self.add_sublattice(sublattice)
        self.aspect = 1/np.sqrt(2)
        self.sublattice.aspect = np.sqrt(2)

        self.sublattice.lattice_vectors = np.array([
            [1, 0],
            [0, 1],
        ])
        self.plot_scale = np.array([np.sqrt(3), np.sqrt(3)/2])

        self.basis = np.zeros((1, 2, self.n_orbitals, 2))
        self.basis[0, 0, :] = (0, 0)
        self.basis[0, 1, :] = (0, 1)
        self.basis = self.basis.reshape(1*2*self.n_orbitals, 2)
        self.gen_coords()

    def init(self):
        match self.sublattice:
            case Square():
                ts = []
                hops = []
                ij = []

                mu = Symbol("mu")
                t = Symbol("t")

                for i in range(self.n_orbitals):
                    ts.append(mu)
                    hops.append([0, 0])
                    ij.append([0, 0, 0, 0, i, i])

                    ts.append(t)
                    hops.append([0, 1])
                    ij.append([0, 1, 0, 0, i, i])

                    ts.append(t)
                    hops.append([-1, 1])
                    ij.append([0, 1, 0, 0, i, i])

                    ts.append(t)
                    hops.append([0, -1])
                    ij.append([0, 0, 0, 1, i, i])

                    ts.append(t)
                    hops.append([1, -1])
                    ij.append([0, 0, 0, 1, i, i])
            case _:
                raise ValueError("Bad sublattice")

        self.ts = [ts, []]
        self.hops = [hops, []]
        self.ij = [ij, []]

    def gen_coords(self):
        a1, a2 = np.sqrt(3) * self.lattice_vectors

        coords_a = np.array([
            x * a1 + y * a2 for x, y in np.ndindex(self.N1, self.N2)
        ]) + self.basis[0]
        coords_b = np.array([
            x * a1 + y * a2 for x, y in np.ndindex(self.N1, self.N2)
        ]) + self.basis[1]

        coords = np.zeros((self.N1, 2, self.N2, 2))
        coords[:, 0] = coords_a.reshape(self.N1, self.N2, 2)
        coords[:, 1] = coords_b.reshape(self.N1, self.N2, 2)
        self.coords = coords.reshape(2*self.N1*self.N2, 2)



from enum import Enum
import warnings
import numpy as np
from keino import Params
from .lattice import BravaisLattice, Space, Site, Symbol
from .impurity import Impurity

class Shape(Enum):
    Chain = "chain"
    Circle = "circle"
    Rectangle = "rectangle"
    Custom = "custom"

"""
chains in either brdige sites or hollow - islands it gets too confusing
same with spacings
assume self similar sublattices

note: for pbc might need to manually move the chain around
"""

class Geometry():
    def __init__(self, N1: int, N2: int, N1_sub: int, N2_sub: int, shape: Shape, lattice: BravaisLattice, params: Params, space: Space, spinful: bool, sites=None, coords=None):
        self.N1 = N1
        self.N2 = N2
        self.N1_sub = N1_sub
        self.N2_sub = N2_sub

        self.shape = Shape(shape)
        self.lattice = BravaisLattice(lattice)
        self.space = Space(space)
        self.spinful = spinful
        self.params = params.copy()

        if self.shape == Shape.Custom:
            assert sites is not None and coords is not None
            self.sites = sites
            self.coords = coords

        self.n_orbitals = 2 if spinful else 1

    def __call__(self, debug=False):
        match self.lattice:
            case BravaisLattice.Square:
                sites, coords, substrate_neighbours, impurity_neighbours = self.square()
            case BravaisLattice.Triangular:
                sites, coords, substrate_neighbours, impurity_neighbours = self.triangular()
            case _:
                raise ValueError(f"Lattice {self.lattice} is not supported!")

        num_impurities = len(sites)

        if debug:
            return sites, coords, substrate_neighbours, impurity_neighbours

        def is_in(idx, lst):
            for val in lst:
                if val == idx:
                    return True
            return False

        impurities = []
        idxs = {}
        t = Symbol("impurities.t")
        Gamma = Symbol("impurities.Gamma")
        mu = Symbol("impurities.mu")
        J = Symbol("impurities.J")
        match self.space:
            case Space.KSpace:
                for n in range(num_impurities):
                    x, y = sites[n]
                    x, x_sub = x // self.N1_sub, x % self.N1_sub
                    y, y_sub = y // self.N2_sub, y % self.N2_sub

                    idx = len(idxs)
                    idxs[(x, y, x_sub, y_sub)] = idx
                    imp = Impurity(idx, self.N1, self.N2, self.N1_sub, self.N2_sub, spinful=self.spinful, coord=coords[n])
                    impurities.append(imp)

                    if self.n_orbitals == 1:
                        imp.add_hopping(imp, 0, 0, -1 * mu)
                    elif self.n_orbitals == 2:
                        imp.add_hopping(imp, 0, 0, (-1 * mu) + J)
                        imp.add_hopping(imp, 1, 1, (-1 * mu) + (-1 * J))

                    for i in range(self.n_orbitals):
                        for x, y, x_sub, y_sub, d1, d2 in substrate_neighbours[n]:
                            imp.add_hopping(Site(x, y, x_sub, y_sub), i, i, Gamma, d1=d1, d2=d2)
                for n in range(num_impurities):
                    imp0 = impurities[n]
                    for x, y, x_sub, y_sub, d1, d2 in impurity_neighbours[n]:
                        if is_in((x, y, x_sub, y_sub), idxs):
                            idx = idxs[(x, y, x_sub, y_sub)]
                            imp1 = impurities[idx]
                            for i in range(self.n_orbitals):
                                imp0.add_hopping(imp1, i, i, t, d1=d1, d2=d2)
            case Space.RealSpace:
                for n in range(num_impurities):
                    x, y = sites[n]
                    x, x_sub = x // self.N1_sub, x % self.N1_sub
                    y, y_sub = y // self.N2_sub, y % self.N2_sub

                    idx = len(idxs)
                    idxs[(x, y, x_sub, y_sub)] = idx
                    imp = Impurity(idx, self.N1, self.N2, self.N1_sub, self.N2_sub, spinful=self.spinful, coord=coords[n])
                    impurities.append(imp)

                    if self.n_orbitals == 1:
                        imp.add_hopping(imp, 0, 0, -1 * mu)
                    elif self.n_orbitals == 2:
                        imp.add_hopping(imp, 0, 0, (-1 * mu) + J)
                        imp.add_hopping(imp, 1, 1, (-1 * mu) + (-1 * J))

                    if len(substrate_neighbours) >= n:
                        for x, y, x_sub, y_sub, d1, d2 in substrate_neighbours[n]:
                            for i in range(self.n_orbitals):
                                imp.add_hopping(Site(x, y, x_sub, y_sub), i, i, Gamma)
                for n in range(num_impurities):
                    imp0 = impurities[n]
                    for x, y, x_sub, y_sub, d1, d2 in impurity_neighbours[n]:
                        if is_in((x, y, x_sub, y_sub), idxs):
                            idx = idxs[(x, y, x_sub, y_sub)]
                            imp1 = impurities[idx]
                            for i in range(self.n_orbitals):
                                imp0.add_hopping(imp1, i, i, 0.5 * t)
            case _:
                raise ValueError(f"Space {self.space} is not supported!")

        return impurities

    def square(self, sites=None):
        pbc = self.params.get("pbc", False)
        if self.space == Space.KSpace:
            pbc = True
        impurity_neighbours = []
        substrate_neighbours = []
        N1 = self.N1 * self.N1_sub
        N2 = self.N2 * self.N2_sub

        match self.shape:
            case Shape.Chain:
                location = self.params.get("location", "hollow")
                spacing = self.params.get("spacing", 0)
                orientation = self.params.get("orientation", "a1")
                L = self.params["L"]
                assert spacing >= 0

                if orientation == "a1":
                    if pbc:
                        if L * (spacing + 1) > N1:
                            warnings.warn("Chain is too long!")
                    else:
                        if L * (spacing + 1) >= N1:
                            warnings.warn("Chain is too long!")
                if orientation == "a2":
                    if pbc:
                        if L * (spacing + 1) > N2:
                            warnings.warn("Chain is too long!")
                    else:
                        if L * (spacing + 1) >= N2:
                            warnings.warn("Chain is too long!")

                x = N1//2 if N1 % 2 else N1//2-1
                y = N2//2 if N2 % 2 else N2//2-1
                centre = self.params.get("centre", (x, y)) # The chain centre is defined in terms of the total number of sites

                match orientation:
                    case "a1":
                        left = centre[0] - (spacing + 1) * L // 2
                        right = centre[0] + (spacing + 1) * L // 2
                        if L % 2 and not spacing % 2:
                            right += 1
                        if left == right:
                            right += 1
                        if L * (spacing + 1) == N1 and pbc:
                            left = 0
                            right = N1
                        sites = [[x, centre[1]] for x in np.arange(left, right, spacing+1)]

                        for n, (X0, Y0) in enumerate(sites):
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = []

                            if X0-(spacing+1) > 0 or pbc:
                                X1, Y1 = (X0-(spacing+1))%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = -1 if pbc else x1-x0
                                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

                            if X0+(spacing+1) < L or pbc:
                                X1, Y1 = (X0+(spacing+1))%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = +1 if pbc else x1-x0
                                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

                            impurity_neighbours.append(neighbours)
                    case "a2":
                        left = centre[1] - (spacing + 1) * L // 2
                        right = centre[1] + (spacing + 1) * L // 2
                        if L % 2:
                            right += 1
                        if left == right == 0:
                            right = 1
                        if L * (spacing + 1) == N2 and pbc:
                            left = 0
                            right = N2
                        sites = [[centre[0], y] for y in np.arange(left, right, spacing+1)]

                        for n, (X0, Y0) in enumerate(sites):
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = []

                            if Y0-(spacing+1) > 0 or pbc:
                                X1, Y1 = X0, (Y0-(spacing+1))%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d2 = -1 if pbc else y1-y0
                                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

                            if Y0+(spacing+1) < L or pbc:
                                X1, Y1 = X0, (Y0+(spacing+1))%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d2 = +1 if pbc else y1-y0
                                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

                            impurity_neighbours.append(neighbours)
                    case _:
                        raise ValueError(f"Orientation {orientation} is not supported!")
                sites = np.array(sites)

                match location:
                    case "hollow":
                        for n in range(L):
                            X0, Y0 = sites[n]
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = [
                                [x0, y0, x0_sub, y0_sub, 0, 0],
                            ]

                            if Y0 + 1 < N2 or pbc:
                                X1, Y1 = X0, (Y0+1)%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = 0
                                d2 = +1 if pbc else y1-y0
                                neighbours.append([x1, y1%self.N2, x1_sub, y1_sub, d1, d2])

                            if X0 + 1 < N1 or pbc:
                                X1, Y1 = (X0+1)%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = +1 if pbc else x1-x0
                                d2 = 0
                                neighbours.append([x1%self.N1, y1, x1_sub, y1_sub, d1, d2])

                            if (X0 + 1 <  N1 and Y0 + 1 < N2) or pbc:
                                X1, Y1 = (X0+1)%N1, (Y0+1)%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = +1 if pbc else x1-x0
                                d2 = +1 if pbc else y1-y0
                                neighbours.append([x1%self.N1, y1%self.N2, x1_sub, y1_sub, d1, d2])

                            substrate_neighbours.append(neighbours)

                        coords = sites + (0.5, 0.5)
                    case "bridge":
                        match orientation:
                            case "a1":
                                for n in range(L):
                                    X0, Y0 = sites[n]
                                    x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                                    y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                                    neighbours = [
                                        [x0, y0, x0_sub, y0_sub, 0, 0],
                                    ]

                                    if X0 + 1 < N1 or pbc:
                                        X1, Y1 = (X0+1)%N1, Y0
                                        x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                        y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                        d1 = +1 if pbc else x1-x0
                                        d2 = 0
                                        neighbours.append([x1, y1, x1_sub, y1_sub, d1, d2])

                                        substrate_neighbours.append(neighbours)
                                coords = sites + (0.5, 0)
                            case "a2":
                                for n in range(L):
                                    X0, Y0 = sites[n]
                                    x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                                    y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                                    neighbours = [
                                        [x0, y0, x0_sub, y0_sub, 0, 0],
                                    ]

                                    if Y0 + 1 < N2 or pbc:
                                        X1, Y1 = X0, (Y0+1)%N2
                                        x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                        y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                        d1 = 0
                                        d2 = +1 if pbc else y1-y0
                                        neighbours.append([x1, y1, x1_sub, y1_sub, d1, d2])

                                        substrate_neighbours.append(neighbours)
                                coords = sites + (0, 0.5)
                            case _:
                                raise ValueError(f"Orientation {orientation} is not supported!")
                    case _:
                        raise ValueError(f"Location {location} is not supported!")
                return sites, coords, substrate_neighbours, impurity_neighbours
            case Shape.Rectangle:
                L = self.params["L"]
                W = self.params["W"]

                if pbc:
                    if L > N2:
                        warnings.warn("Island is too long!")
                else:
                    if L >= N2:
                        warnings.warn("Island is too long!")
                if pbc:
                    if W > N1:
                        warnings.warn("Island is too wide!")
                else:
                    if W >= N2:
                        warnings.warn("Island is too wide!")

                x = N1//2 if N1 % 2 else N1//2-1
                y = N2//2 if N2 % 2 else N2//2-1

                centre = self.params.get("centre", (x, y)) # The island centre is defined in terms of the total number of sites

                left = centre[0] - W // 2
                right = centre[0] + W // 2
                bottom = centre[1] - L // 2
                top = centre[1] + L // 2

                if W % 2:
                    right += 1
                if L % 2:
                    top += 1
                if left == right:
                    right += 1
                if top == bottom:
                    top += 1

                if W == N1 and pbc:
                    left = 0
                    right = N1
                if L == N2 and pbc:
                    bottom = 0
                    top = N2

                sites = np.meshgrid(np.arange(left, right), np.arange(bottom, top))
                sites = np.array(sites).reshape(2, L*W).T
            case Shape.Circle:
                pbc = False
                R = self.params["R"]

                assert 2 * R < N1 and 2 * R < N2

                x = N1//2 if N1 % 2 else N1//2-1
                y = N2//2 if N2 % 2 else N2//2-1

                centre = self.params.get("centre", (x, y)) # The island centre is defined in terms of the total number of sites

                sites = []
                for x, y in np.ndindex(N1, N2):
                    if np.sqrt((x - centre[0])**2 + (y - centre[1])**2) <= R:
                        sites.append([x, y])
                sites = np.array(sites)
            case Shape.Custom:
                sites = self.sites
                coords = self.coords
            case _:
                raise ValueError(f"Shape {self.shape} is not supported!")

        match self.shape:
            case Shape.Rectangle:
                if self.space == Space.RealSpace:
                    pbc_x = pbc and L > 1
                    pbc_y = pbc and W > 1
                else:
                    pbc_x = True
                    pbc_y = True
            case Shape.Circle:
                pbc_x = pbc
                pbc_y = pbc
                L = 2*R
                W = 2*R
            case Shape.Custom:
                pbc_x = False
                pbc_y = False
                L = N1
                W = N2
            case _:
                raise ValueError(f"Shape {self.shape} is not supported!")
        for n, (X0, Y0) in enumerate(sites):
            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

            neighbours = []

            if X0-1 > 0 or pbc_x:
                X1, Y1 = (X0-1)%N1, Y0
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d1 = -1 if pbc else x1-x0
                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

            if X0+1 < L or pbc_x:
                X1, Y1 = (X0+1)%N1, Y0
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d1 = +1 if pbc else x1-x0
                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

            if Y0-1 > 0 or pbc_y:
                X1, Y1 = X0, (Y0-1)%N2
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d2 = -1 if pbc else y1-y0
                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

            if Y0+1 < W or pbc_y:
                X1, Y1 = X0, (Y0+1)%N2
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d2 = +1 if pbc else y1-y0
                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

            impurity_neighbours.append(neighbours)

        for n, (X0, Y0) in enumerate(sites):
            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

            neighbours = [
                [x0, y0, x0_sub, y0_sub, 0, 0],
            ]

            if Y0 + 1 < N2 or pbc:
                X1, Y1 = X0, (Y0+1)%N2
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d1 = 0 if pbc else x1-x0
                d2 = +1 if pbc else y1-y0
                neighbours.append([x1, y1%self.N2, x1_sub, y1_sub, d1, d2])

            if X0 + 1 < N1 or pbc:
                X1, Y1 = (X0+1)%N1, Y0
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d1 = +1 if pbc else x1-x0
                d2 = 0 if pbc else y1-y0
                neighbours.append([x1%self.N1, y1, x1_sub, y1_sub, d1, d2])

            if (X0 + 1 <  N1 and Y0 + 1 < N2) or pbc:
                X1, Y1 = (X0+1)%N1, (Y0+1)%N2
                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                d1 = +1 if pbc else x1-x0
                d2 = +1 if pbc else y1-y0
                neighbours.append([x1%self.N1, y1%self.N2, x1_sub, y1_sub, d1, d2])

            substrate_neighbours.append(neighbours)

        coords = sites + (0.5, 0.5)

        return sites, coords, substrate_neighbours, impurity_neighbours

    def triangular(self, sites=None):
        pbc = self.params.get("pbc", False)
        if self.space == Space.KSpace:
            pbc = True
        impurity_neighbours = []
        substrate_neighbours = []
        N1 = self.N1 * self.N1_sub
        N2 = self.N2 * self.N2_sub

        match self.shape:
            case Shape.Chain:
                location = self.params.get("location", "hollow")
                spacing = self.params.get("spacing", 0)
                orientation = self.params.get("orientation", "a1")
                L = self.params["L"]
                assert spacing >= 0

                if orientation == "a1":
                    if pbc:
                        if L * (spacing + 1) > N1:
                            warnings.warn("Chain is too long!")
                    else:
                        if L * (spacing + 1) >= N1:
                            warnings.warn("Chain is too long!")
                if orientation == "a2":
                    if pbc:
                        if L * (spacing + 1) > N2:
                            warnings.warn("Chain is too long!")
                    else:
                        if L * (spacing + 1) >= N2:
                            warnings.warn("Chain is too long!")

                x = N1//2 if N1 % 2 else N1//2-1
                y = N2//2 if N2 % 2 else N2//2-1
                centre = self.params.get("centre", (x, y)) # The chain centre is defined in terms of the total number of sites

                match orientation:
                    case "a1":
                        left = centre[0] - (spacing + 1) * L // 2
                        right = centre[0] + (spacing + 1) * L // 2
                        if L % 2 and not spacing % 2:
                            right += 1
                        if left == right:
                            right += 1
                        if L * (spacing + 1) == N1 and pbc:
                            left = 0
                            right = N1
                        sites = [[x, centre[1]] for x in np.arange(left, right, spacing+1)]

                        for n, (X0, Y0) in enumerate(sites):
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = []

                            if X0-(spacing+1) > 0 or pbc:
                                X1, Y1 = (X0-(spacing+1))%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = -1 if pbc else x1-x0
                                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

                            if X0+(spacing+1) < L or pbc:
                                X1, Y1 = (X0+(spacing+1))%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = +1 if pbc else x1-x0
                                neighbours.append([x1, y1, x1_sub, y1_sub, d1, 0])

                            impurity_neighbours.append(neighbours)
                    case "a2":
                        left = centre[1] - (spacing + 1) * L // 2
                        right = centre[1] + (spacing + 1) * L // 2
                        if L % 2:
                            right += 1
                        if left == right == 0:
                            right = 1
                        if L * (spacing + 1) == N2 and pbc:
                            left = 0
                            right = N2
                        sites = [[centre[0], y] for y in np.arange(left, right, spacing+1)]

                        for n, (X0, Y0) in enumerate(sites):
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = []

                            if Y0-(spacing+1) > 0 or pbc:
                                X1, Y1 = X0, (Y0-(spacing+1))%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d2 = -1 if pbc else y1-y0
                                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

                            if Y0+(spacing+1) < L or pbc:
                                X1, Y1 = X0, (Y0+(spacing+1))%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d2 = +1 if pbc else y1-y0
                                neighbours.append([x1, y1, x1_sub, y1_sub, 0, d2])

                            impurity_neighbours.append(neighbours)
                    case _:
                        raise ValueError(f"Orientation {orientation} is not supported!")
                sites = np.array(sites)

                match location:
                    case "hollow":
                        for n in range(L):
                            X0, Y0 = sites[n]
                            x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                            y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                            neighbours = [
                                [x0, y0, x0_sub, y0_sub, 0, 0],
                            ]

                            if Y0 + 1 < N2 or pbc:
                                X1, Y1 = X0, (Y0+1)%N2
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = 0
                                d2 = +1 if pbc else y1-y0
                                neighbours.append([x1, y1%self.N2, x1_sub, y1_sub, d1, d2])

                            if X0 + 1 < N1 or pbc:
                                X1, Y1 = (X0+1)%N1, Y0
                                x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                d1 = +1 if pbc else x1-x0
                                d2 = 0
                                neighbours.append([x1%self.N1, y1, x1_sub, y1_sub, d1, d2])

                            substrate_neighbours.append(neighbours)

                        coords = sites + (0.25, 0.5)
                    case "bridge":
                        match orientation:
                            case "a1":
                                for n in range(L):
                                    X0, Y0 = sites[n]
                                    x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                                    y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                                    neighbours = [
                                        [x0, y0, x0_sub, y0_sub, 0, 0],
                                    ]

                                    if X0 + 1 < N1 or pbc:
                                        X1, Y1 = (X0+1)%N1, Y0
                                        x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                        y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                        d1 = +1 if pbc else x1-x0
                                        d2 = 0
                                        neighbours.append([x1, y1, x1_sub, y1_sub, d1, d2])

                                        substrate_neighbours.append(neighbours)
                                coords = sites + (0.25, 0)
                            case "a2":
                                for n in range(L):
                                    X0, Y0 = sites[n]
                                    x0, x0_sub = X0 // self.N1_sub, X0 % self.N1_sub
                                    y0, y0_sub = Y0 // self.N2_sub, Y0 % self.N2_sub

                                    neighbours = [
                                        [x0, y0, x0_sub, y0_sub, 0, 0],
                                    ]

                                    if Y0 + 1 < N2 or pbc:
                                        X1, Y1 = X0, (Y0+1)%N2
                                        x1, x1_sub = X1 // self.N1_sub, X1 % self.N1_sub
                                        y1, y1_sub = Y1 // self.N2_sub, Y1 % self.N2_sub
                                        d1 = 0
                                        d2 = +1 if pbc else y1-y0
                                        neighbours.append([x1, y1, x1_sub, y1_sub, d1, d2])

                                        substrate_neighbours.append(neighbours)
                                coords = sites + (0, 0.25)
                            case _:
                                raise ValueError(f"Orientation {orientation} is not supported!")
                    case _:
                        raise ValueError(f"Location {location} is not supported!")
                return sites, coords, substrate_neighbours, impurity_neighbours
            case _:
                raise ValueError(f"Shape {shape} is not implemented!")



# pylint: disable=line-too-long, fixme, too-many-arguments, too-many-positional-arguments, invalid-name, too-many-locals, too-few-public-methods, too-many-instance-attributes

"""
file:     ristikko/lattice.py
Homepage: TODO
Licence:  GPLv3+

Defines a lattice for tight-binding models (Lattice), bravais lattices (BravaisLattice),
different spaces allowing for Fourier transforms (Space), and abstract
symbols for tight-binding Hamiltonians (Symbol).
"""

from dataclasses import dataclass
import copy
from enum import Enum
import numpy as np
from keino import Params, ndindex

@dataclass
class Site():
    """Represents a site on a lattice. (x, y) is the superlattice coordinates and (x_sub, y_sub) the sublattice."""
    x: int
    y: int
    x_sub: int
    y_sub: int

    def __iter__(self):
        yield self.x
        yield self.y
        yield self.x_sub
        yield self.y_sub

# pylint: disable=missing-class-docstring
class BravaisLattice(Enum):
    Triangular = "triangular"
    Square = "square"

    def lattice_vectors(self):
        # pylint: disable=missing-function-docstring
        match self.name:
            case "Triangular":
                lattice_vectors = np.array([
                    [1, 0],
                    [0.5, np.sqrt(3)/2],
                ])
            case "Square":
                lattice_vectors = np.array([
                    [1, 0],
                    [0, 1],
                ])
            case _:
                raise ValueError("Unsupported Bravais lattice")
        return lattice_vectors

# pylint: disable=missing-class-docstring
class Space(Enum):
    KSpace = "kspace"
    RealSpace = "realspace"
    Ribbon = "ribbon"

class Symbol():
    """
    Represents a symbol in a Hamiltonian.

    In ristikko, a lattice is defined by a Bravais lattice, a sublattice,
    the hoppings with the sublattice, and hoppings between sublattices.
    Each sublattice, in turn, has a sublattice.
    Hoppings are defined using Symbols.

    When a Hamiltonian matrix is constructed, the Symbols are replaced
    with a number, drawn from the provided Params object.

    The simplest case is a symbol with a single name:

        t = Symbol("t")

    Symbols can be multiplied by a prefactor

        t = -2 * Symbol("t")

    Symbols can be added together

        f = Symbol("a") + Symbol("b")

    This produces a symbol with the name ["a", "b"].

    In this fashion, arbitrarily complicated symbols can be constructed.
    """

    def __init__(self, names, multipliers=None):
        assert isinstance(names, ( str, list, np.ndarray ))

        self.names = names

        if multipliers is None:
            match names:
                case str():
                    multipliers = 1
                case np.ndarray():
                    multipliers = np.zeros_like(names)
                    for idx in ndindex(*names.shape):
                        if names[idx]:
                            multipliers[idx] = 1
                case list():
                    multipliers = [1 for _ in self.names]
                case _:
                    raise ValueError("names must be str, list, or np.ndarray")

        self.multipliers = multipliers

    def __getitem__(self, index):
        names = self.names[index]
        multipliers = self.multipliers[index]
        new_symbol = Symbol(names)
        new_symbol.multipliers = multipliers
        return new_symbol

    def replace(self, params, x, y, default=0):
        """Replace a symbol with the corresponding value(s) from params"""
        def params_get(key):
            value = params.get(key, default)
            if isinstance(value, np.ndarray):
                return value[x, y]
            return value

        match self.names:
            case str():
                val = params_get(self.names) * self.multipliers
            case np.ndarray():
                val = np.zeros_like(self.names, complex)
                for idx in ndindex(*self.names.shape):
                    val[idx] = params_get(self.names[idx]) * self.multipliers[idx]
                val = np.sum(val)
            case list():
                val = [
                        params_get(names) * multiplier for names, multiplier in zip(self.names, self.multipliers)
                      ]
                val = sum(val)
            case _:
                raise ValueError("names must be str, list, or np.ndarray")

        return val

    def conj(self):
        match self.names:
            case str():
                self.multipliers = complex(np.conj(self.multipliers))
            case np.ndarray():
                self.multipliers = np.conj(self.multipliers)
            case list():
                self.multipliers = np.conj(self.multipliers).tolist()
            case _:
                raise ValueError("names must be str, list, or np.ndarray")

    def __eq__(self, other):
        if isinstance(other, Symbol):
            return self.names == other.names
        return False

    def add(self, other):
        # pylint: disable=too-many-branches
        if other is None:
            return self

        names = None
        multipliers = None
        if isinstance(other, Symbol):
            match (self.names, other.names):
                case (str(), str()):
                    names = [self.names, other.names]
                    multipliers = [self.multipliers, other.multipliers]
                case (str(), np.ndarray()):
                    names = np.empty_like(other.names.shape + (2,), object)
                    multipliers = np.zeros_like(other.names.shape + (2,))
                    for idx in ndindex(*other.names.shape):
                        names[idx] = [other.names[idx], self.names]
                        multipliers[idx] = [other.multipliers[idx], self.multipliers]
                case (str(), list()):
                    names = [self.names] + other.names
                    multipliers = [self.multipliers] + other.multipliers
                case (np.ndarray(), str()):
                    names = np.empty_like(self.names.shape + (2,), object)
                    multipliers = np.zeros_like(self.names.shape + (2,), object)
                    for idx in ndindex(*self.names.shape):
                        names[idx] = [self.names[idx], other.names]
                        multipliers[idx] = [self.multipliers[idx], other.multipliers]
                case (np.ndarray(), np.ndarray()):
                    names = self.names + other.names
                    multipliers = self.multipliers + other.multipliers
                case (np.ndarray(), list()):
                    ...
                case (list(), str()):
                    names = self.names + [other.names]
                    multipliers = self.multipliers + [other.multipliers]
                case (list(), list()):
                    names = self.names + other.names
                    multipliers = self.multipliers + other.multipliers
                case (list(), np.ndarray()):
                    ...
                case _:
                    raise ValueError("self.names and other.names must be either str, list, or np.ndarray!")

        new_symbol = Symbol(names)
        new_symbol.multipliers = multipliers
        return new_symbol

    def __add__(self, other):
        return self.add(other)
    def __radd__(self, other):
        return self.add(other)

    def mul(self, other):
        # pylint: disable=missing-function-docstring, too-many-branches
        names = None
        multipliers = None

        match (self.names, other):
            case (str(), complex() | float() | int()):
                names = self.names
                multipliers = self.multipliers * other
            case (str(), np.ndarray()):
                names = np.empty_like(other, object)
                multipliers = other * self.multipliers
                for idx in ndindex(*names.shape):
                    if multipliers[idx] is not None:
                        names[idx] = self.names
                    else:
                        names[idx] = None
            case (str(), list()):
                names = [self.names for _ in other]
                multipliers = [m * self.multipliers for m in other]
            case (np.ndarray(), complex() | float() | int()):
                names = self.names
                multipliers = self.multipliers * other
            case (np.ndarray(), np.ndarray()):
                names = self.names
                multipliers = self.multipliers * other
            case (np.ndarray(), list()):
                ...
            case (list(), complex() | float() | int()):
                names = self.names
                multipliers = [m * other for m in self.multipliers]
            case (list(), list()):
                names = self.names
                multipliers = [m1 * m2 for m1, m2 in zip(self.multipliers, other)]
            case (list(), np.ndarray()):
                ...

        new_parameter = Symbol(names)
        new_parameter.multipliers = multipliers
        return new_parameter

    def __rmul__(self, other):
        return self.mul(other)

    def __mul__(self, other):
        return self.mul(other)

    def __str__(self):
        if isinstance(self.names, list):
            ret = str([str(l) + " * " + r for l, r in zip(self.multipliers, self.names)])
        else:
            ret = str(self.multipliers) + " * " + self.names

        return str(ret)

    def __repr__(self):
        if isinstance(self.names, list):
            ret = str([str(l) + " * " + r for l, r in zip(self.multipliers, self.names)])
        else:
            ret = str(self.multipliers) + " * " + self.names

        return str(ret)

class Floor():
    """
    In ristikko, a lattice can be constructed out of arbitrarily nested lattices.
    Floot represents the bottom of this nest.
    """
    def __init__(self):
        self.N1 = 1
        self.N2 = 1
        self._hoppings = None
        self._sc_hoppings = None
        self.removed_sites = []

    def set_n_orbitals(self, n_orbitals):
        pass

    def __call__(self, params, space, superconducting, **kwargs):
        return [], []

def _gen_hoppings_realspace(n_orbitals, N1_super, N2_super, N1_sub, N2_sub, pbc, _ts, _hops, _ij, _ts_sub, _ij_sub, valid_sites, removed_energy=10):
    #pylint: disable=consider-using-enumerate
    """Generates a list of hoppings for a Hamiltonian in real space"""
    ts = []
    ij = []

    ind = np.arange(N1_sub*N2_sub*n_orbitals).reshape(N1_sub, N2_sub, n_orbitals)

    # Intercell hops
    for n, x, y in np.ndindex(len(_ts), N1_super, N2_super):
        x_sub0, y_sub0, x_sub1, y_sub1, i, j = _ij[n]
        t = _ts[n]
        d1, d2 = _hops[n]

        ii = ind[x_sub0, y_sub0, i]
        jj = ind[x_sub1, y_sub1, j]
        x0, y0 = x, y
        x1, y1 = (x+d1)%N1_super, (y+d2)%N2_super

        if valid_sites[x0, y0, x_sub0, y_sub0] or valid_sites[x1, y1, x_sub1, y_sub1]:
            continue

        test_x = True if pbc[0] else 0 <= x+d1 < N1_super
        test_y = True if pbc[1] else 0 <= y+d2 < N2_super
        if test_x and test_y:
            ts.append(t)
            ij.append([x0, y0, x1, y1, ii, jj])

    # Intracell hops
    for n, x, y in np.ndindex(len(_ts_sub), N1_super, N2_super):
        x0, y0, x1, y1, i, j = _ij_sub[n]
        t = _ts_sub[n]
        ii = ind[x0, y0, i]
        jj = ind[x1, y1, j]

        if valid_sites[x, y, x0, y0] or valid_sites[x, y, x1, y1]:
            if x0 == x1 and y0 == y1 and i == j:
                t = removed_energy
            else:
                if t != 0 and t is not None:
                    t = 0

        ts.append(t)
        ij.append([x, y, x, y, ii, jj])

    return ts, ij

def _gen_hoppings_ribbon(k, lattice_vectors, n_orbitals, L, direction, pbc, _ts, _hops, _ij, _ts_sub, _ij_sub):
    # pylint: disable=consider-using-enumerate
    """Generates a list of hoppings for a Ribbon Hamiltonian"""
    ts = []
    ij = []

    ind = np.arange(L*n_orbitals).reshape(L, n_orbitals)

    # Intercell hops
    if direction == "a1":
        for n in range(len(_ts)):
            for y in range(L):
                t = _ts[n]
                x0, y0, x1, y1, i, j = _ij[n]
                d1, d2 = _hops[n]

                vec = (d1 * lattice_vectors[0]) + 0.0j
                t *= np.exp(1j * k @ vec)

                ii = ind[y0, i]
                jj = ind[y1, j]

                test = True if pbc else 0 <= y+d2 < L
                if test:
                    ts.append(t)
                    ij.append([y, (y+d2)%L, ii, jj])
    elif direction == "a2":
        for n in range(len(_ts)):
            for x in range(L):
                t = _ts[n]
                x0, y0, x1, y1, i, j = _ij[n]
                d1, d2 = _hops[n]

                vec = (d2 * lattice_vectors[1]) + 0.0j
                t *= np.exp(1j * k @ vec)

                ii = ind[x0, i]
                jj = ind[x1, j]

                test = True if pbc else 0 <= x+d1 < L
                if test:
                    ts.append(t)
                    ij.append([x, (x+d1)%L, ii, jj])

    # Intracell hops
    for n in range(len(_ts_sub)):
        x0, y0, x1, y1, i, j = _ij_sub[n]
        t = _ts_sub[n]
        ii = ind[x0, i]
        jj = ind[x1, j]

        for x in range(L):
            ts.append(t)
            ij.append([x, x, ii, jj])

    return ts, ij


class Lattice():
    """
    Represents a lattice.

    In ristikko, a lattice is defined by a Bravais lattice, a sublattice,
    the hoppings with the sublattice, and hoppings between sublattices.
    Hoppings are defined using Symbols.
    Each sublattice, in turn, has a sublattice.

    Given some parameters, and whether or not to Fourier transform,
    this class will generate a list of hoppings for a Hamiltonian.
    """
    def __init__(self, bravais_lattice: BravaisLattice, coords=None, hopping_degree=1, aspect=1, **kwargs):
        assert bravais_lattice in BravaisLattice

        self.basis = None
        self.n_orbitals = None
        self.bravais_lattice = bravais_lattice
        self.lattice_vectors = bravais_lattice.lattice_vectors()
        self.sublattice = None
        self.hd = hopping_degree
        self.coords = coords
        self.aspect = aspect
        self.plot_scale = 1

        self.ts = None
        self.ij = None
        self.hops = None

        sigma = np.array([
            [[0, 1], [1, 0]],
            [[0, -1j], [1j, 0]],
        ])
        self.rashba_matrix = 1j * (np.outer(self.lattice_vectors[:, 0], sigma[1]) + np.outer(self.lattice_vectors[:, 1], sigma[0])).reshape(2, 2, 2)
        self.dresselhaus_matrix = 1j * (np.outer(self.lattice_vectors[:, 0], sigma[0]) + np.outer(self.lattice_vectors[:, 1], sigma[1])).reshape(2, 2, 2)

        self.N1 = kwargs.get("N1", 1)
        self.N2 = kwargs.get("N2", 1)
        self.L = kwargs.get("L", 1)

        self.removed_sites = []


    def init(self):
        pass

    def copy(self):
        return copy.deepcopy(self)

    def set_n_orbitals(self, n_orbitals):
        self.n_orbitals = n_orbitals

    def add_sublattice(self, sublattice):
        # pylint: disable=attribute-defined-outside-init
        assert isinstance(sublattice, Lattice)

        self.sublattice = sublattice
        self.sublattice.set_n_orbitals(self.n_orbitals)
        self.aspect = self.sublattice.aspect
        if self.sublattice.basis is not None:
            self.basis = self.sublattice.basis

    def shape(self):
        if self.sublattice is None:
            # pylint: disable=import-outside-toplevel
            from .models import Point
            point = Point(self.bravais_lattice)
            self.add_sublattice(point)
        return self.N1, self.N2, self.sublattice.N1, self.sublattice.N2

    def preprocess_hoppings(self):
        pass

    def gen_coords(self):
        if self.coords is None:
            N1_super, N2_super, N1_sub, N2_sub = self.shape()
            a1 = self.lattice_vectors[0]
            a2 = self.lattice_vectors[1]
            self.coords = np.array([
                x * a1 + y * a2 for x, y in np.ndindex(N1_sub*N1_super, N2_sub*N2_super)
            ])

    def remove_site(self, site: Site):
        self.removed_sites.append(site)

    def __call__(self, params: Params, space: Space, superconducting=False, **kwargs):
        # pylint: disable=inconsistent-return-statements
        """Generate a list of hoppings"""

        if self.sublattice is None:
            # pylint: disable=import-outside-toplevel
            from .models import Point
            point = Point(self.bravais_lattice)
            self.add_sublattice(point)
        self.sublattice.set_n_orbitals(self.n_orbitals)

        if self.basis is None:
            self.basis = np.zeros((self.sublattice.N1, self.sublattice.N2, self.n_orbitals, 2))
            a1, a2 = self.lattice_vectors[0], self.lattice_vectors[1]
            for x, y in np.ndindex(self.sublattice.N1, self.sublattice.N2):
                coord = x * a1 + y * a2
                self.basis[x, y, :] = coord
            self.basis = self.basis.reshape(self.n_orbitals*self.sublattice.N1*self.sublattice.N2, 2)

        self.init()
        self.preprocess_hoppings()
        self.gen_coords()

        match space:
            case Space.RealSpace:
                return self.gen_hoppings_realspace(params, space, superconducting, **kwargs)
            case Space.KSpace:
                return self.gen_hoppings_kspace(params, space, superconducting, **kwargs)
            case Space.Ribbon:
                return self.gen_hoppings_ribbon(params, space, superconducting, **kwargs)
            case _:
                raise ValueError(f"Space {space} not implemented")

    def gen_hoppings_realspace(self, params, space, superconducting, impurities=None, replace_symbols=True, is_sublattice=False, trim=False):
        # pylint: disable=unsubscriptable-object, too-many-branches
        N1_super, N2_super, N1_sub, N2_sub = self.shape()

        _ts, _hops, _ij = self.ts[0], self.hops[0], self.ij[0]
        if superconducting:
            _ts, _hops, _ij = self.ts[1], self.hops[1], self.ij[1]

        _ts_sub, _ij_sub = self.sublattice(params, space, superconducting, replace_symbols=replace_symbols, is_sublattice=True)

        if is_sublattice:
            pbc = [False, False]
        else:
            if hasattr(params.pbc, "__len__"):
                pbc = params.pbc
            else:
                pbc = [params.pbc, params.pbc]
            pbc = np.array(pbc)

        valid_sites = np.zeros((N1_super, N2_super, N1_sub, N2_sub))
        for site in self.removed_sites:
            valid_sites[site.x, site.y, site.x_sub, site.y_sub] = 1

        removed_energy = params.get("removed_energy", 10)

        ts, ij = _gen_hoppings_realspace(self.n_orbitals, N1_super, N2_super, N1_sub, N2_sub, pbc, _ts, _hops, _ij, _ts_sub, _ij_sub, valid_sites, removed_energy=removed_energy)

        if impurities is not None:
            for impurity in impurities:
                _ts = impurity(params, space)
                for t, hop in zip(_ts, impurity.hoppings):
                    ts.append(t)
                    ij.append([impurity.x, impurity.y, hop.other.x, hop.other.y, hop.i, hop.j])
                    if hop.i != hop.j:
                        ts.append(t)
                        ij.append([hop.other.x, hop.other.y, impurity.x, impurity.y, hop.j, hop.i])

        if replace_symbols:
            for n, t in enumerate(ts):
                x0, y0, _, _, _, _ = ij[n]
                if isinstance(t, Symbol):
                    ts[n] = t.replace(params, x0, y0)

            ts = np.array(ts, dtype=complex)
            ij = np.array(ij, dtype=int)
        else:
            ts = np.array(ts, dtype=object)
            ij = np.array(ij, dtype=int)

        if trim:
            trim = []
            for n in range(ts.size):
                if ts[n] == 0 or ts[n] == removed_energy:
                    trim.append(n)

            ts = np.delete(ts, trim, 0)
            ij = np.delete(ij, trim, 0)

        return ts, ij

    def gen_hoppings_kspace(self, params, space, superconducting, k=None, direction=None, impurities=None):
        # pylint: disable=too-many-branches, unsubscriptable-object, consider-using-enumerate, too-many-statements
        if direction is not None:
            assert direction in ["a1", "a2", "both"]
        else:
            direction = "both"

        k = np.array(k)
        assert 0 < k.size < 3
        if k.size == 1:
            if direction == "a1":
                k = np.append(k, 0)
            elif direction == "a2":
                k = np.append(0, k)
            else:
                raise ValueError("Momenta must be 2D!")

        _ts, _hops, _ij = self.ts[0].copy(), self.hops[0].copy(), self.ij[0].copy()
        if superconducting:
            _ts, _hops, _ij = self.ts[1].copy(), self.hops[1].copy(), self.ij[1].copy()

        _ts_sub, _ij_sub = self.sublattice(params, Space.RealSpace, superconducting, is_sublattice=True)

        for n in range(len(_ts)):
            x, y, _, _, _, _ = _ij[n]
            _ts[n] = _ts[n].replace(params, x, y)

        hops = []
        ij = []
        ts = []

        ind = np.arange(self.sublattice.N1*self.sublattice.N2*self.n_orbitals).reshape(self.sublattice.N1, self.sublattice.N2, self.n_orbitals)

        # Intercell hops
        for n, t in enumerate(_ts):
            x0, y0, x1, y1, i, j = _ij[n]
            d1, d2 = _hops[n]
            t *= np.exp(1j * k @ (d1 * self.lattice_vectors[0] + d2 * self.lattice_vectors[1]))

            ii = ind[x0, y0, i]
            jj = ind[x1, y1, j]

            if t != 0:
                hops.append([d1, d2])
                ij.append([ii, jj])
                ts.append(t)

        # Intracell hops
        for n, t in enumerate(_ts_sub):
            x0, y0, x1, y1, i, j = _ij_sub[n]

            ii = ind[x0, y0, i]
            jj = ind[x1, y1, j]

            if t != 0:
                hops.append([0, 0])
                ij.append([ii, jj])
                ts.append(t)

        hops = np.array(hops, int)
        ij = np.array(ij, int)
        ts = np.array(ts, complex)

        if direction == "a1":
            for d2 in range(-self.hd, self.hd+1):
                if d2 != 0:
                    ts[hops[:, 1] == d2] = 0
        elif direction == "a2":
            for d1 in range(-self.hd, self.hd+1):
                if d1 != 0:
                    ts[hops[:, 0] == d1] = 0

        if impurities is not None:
            for impurity in impurities:
                _ts = impurity(params, space, lattice_vectors=self.lattice_vectors, k=k)
                for t, hop in zip(_ts, impurity.hoppings):
                    ts = np.append(ts, t)
                    ij = np.append(ij, [[hop.i, hop.j]], axis=0)
                    if hop.i != hop.j:
                        ij = np.append(ij, [[hop.j, hop.i]], axis=0)
                        ts = np.append(ts, np.conj(t))

        trim = []
        for n in range(ts.size):
            if ts[n] == 0:
                trim.append(n)

        ts = np.delete(ts, trim, 0)
        ij = np.delete(ij, trim, 0)

        return ts, ij

    def gen_hoppings_ribbon(self, params, space, superconducting, k=None, direction=None, impurities=None):
        # pylint: disable=too-many-branches, unsubscriptable-object, consider-using-enumerate
        if direction is not None:
            assert direction in ["a1", "a2"]
        else:
            direction = "a1"

        k = np.array(k)
        assert 0 < k.size < 3
        if k.size == 1:
            if direction == "a1":
                k = np.append(k, 0)
            elif direction == "a2":
                k = np.append(0, k)
            else:
                raise ValueError("Must Fourier transform along either a1 or a2!")

        _ts, _hops, _ij = self.ts[0], self.hops[0], self.ij[0]
        if superconducting:
            _ts, _hops, _ij = self.ts[1], self.hops[1], self.ij[1]

        _ts_sub, _ij_sub = self.sublattice(params, Space.RealSpace, superconducting, is_sublattice=True)

        for n in range(len(_ts)):
            x, y, _, _, _, _ = _ij[n]
            if isinstance(_ts[n], Symbol):
                _ts[n] = _ts[n].replace(params, x, y)
        for n in range(len(_ts_sub)):
            x, y, _, _, _, _ = _ij_sub[n]
            if isinstance(_ts_sub[n], Symbol):
                _ts_sub[n] = _ts_sub[n].replace(params, x, y)

        pbc = params.pbc

        _ts = np.array(_ts, dtype=complex)
        _hops = np.array(_hops, dtype=int)
        _ij = np.array(_ij, dtype=int)
        _ts_sub = np.array(_ts_sub, dtype=complex)
        _ij_sub = np.array(_ij_sub, dtype=int)

        ts, ij = _gen_hoppings_ribbon(k, self.lattice_vectors, self.n_orbitals, self.L, direction, pbc, _ts, _hops, _ij, _ts_sub, _ij_sub)

        if impurities is not None:
            if direction == "a1":
                for impurity in impurities:
                    _ts = impurity(params, space, direction=direction, lattice_vectors=self.lattice_vectors, k=k)

                    for t, hop in zip(_ts, impurity.hoppings):
                        ts.append(t)
                        ij.append([impurity.y, hop.other.y, hop.i, hop.j])
                        if hop.i != hop.j:
                            ts.append(t)
                            ij.append([hop.other.y, impurity.y, hop.j, hop.i])
            elif direction == "a2":
                for impurity in impurities:
                    _ts = impurity(params, space, direction=direction, lattice_vectors=self.lattice_vectors, k=k)

                    for t, hop in zip(_ts, impurity.hoppings):
                        ts.append(t)
                        ij.append([impurity.x, hop.other.x, hop.i, hop.j])
                        if hop.i != hop.j:
                            ts.append(t)
                            ij.append([hop.other.x, impurity.x, hop.j, hop.i])

        ts = np.array(ts, dtype=complex)
        ij = np.array(ij, dtype=int)

        return ts, ij



# pylint: disable=invalid-name, line-too-long
"""Sublattice with a single site"""
import numpy as np
from ..lattice import Lattice, BravaisLattice, Symbol, Floor

class Point(Lattice):
    def __init__(self, bravais_lattice: BravaisLattice):
        self.bravais_lattice = bravais_lattice
        self.N1 = 1
        self.N2 = 1

        aspect = 1
        if bravais_lattice == BravaisLattice.Triangular:
            aspect = 1/np.sqrt(3)

        super().__init__(bravais_lattice, aspect=aspect)

        self.sublattice = Floor()

    def init(self):
        mu = Symbol("mu")
        Jx = Symbol("Jx")
        Jy = Symbol("Jy")
        Jz = Symbol("Jz")

        if self.n_orbitals == 2:
            ts = [
                ( -1 * mu ) + Jz,
                ( -1 * mu ) + ( -1 * Jz ),
                Jx  + (- 1j * Jy ),
                Jx + 1j * Jy
            ]
            ij = [
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1, 0],
            ]
            hops = [
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
            ]

            sc_ts = [Symbol("Delta.s.0"), Symbol("Delta.s.0", -1)]
            sc_ij = [
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1, 0],
            ]

            sc_hops = [
                [0, 0],
                [0, 0],
            ]
        else:
            sc_ts = []
            sc_ij = []
            sc_hops = []
            ts = [-1 * mu]
            ij = [
                [0, 0, 0, 0, 0, 0]
            ]
            hops = [[0, 0]]

        self.ts = [ts, sc_ts]
        self.hops = [hops, sc_hops]
        self.ij = [ij, sc_ij]



#pylint: disable=missing-module-docstring

from .lattice import Lattice, BravaisLattice, Space, Symbol, Site
from .system import System
from .impurity import Impurity, Hopping
from .geometry import Geometry

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("ristikko")
except PackageNotFoundError:
    # package is not installed
    pass



# pylint: disable=invalid-name
"""
Numerical calculation of the Chern number using the Fukui-Hatsugai method
"""
import numpy as np

### FHS method
def bump_x(k, Nk):
    """Given the Brillouin zone is discretised into Nk points,
    increment a k-index pair in the x-direction, handling periodic
    boundaries"""
    return ( (k[0] + 1) % Nk, k[1] )
def bump_y(k, Nk):
    """Given the Brillouin zone is discretised into Nk points,
    increment a k-index pair in the y-direction, handling periodic
    boundaries"""
    return ( k[0], (k[1] + 1) % Nk )

def link1(k, band, Nk):
    """FHS U(1) link variable in the x-direction"""
    k_n = k
    k_m = bump_x(k, Nk)
    n = band[k_n]
    m = band[k_m]

    numer = np.vdot(n, m)
    denom = np.linalg.norm(numer)
    U = numer / denom
    return U

def link2(k, band, Nk):
    """FHS U(1) link variable in the y-direction"""
    k_n = k
    k_m = bump_y(k, Nk)
    n = band[k_n]
    m = band[k_m]

    numer = np.vdot(n, m)
    denom = np.linalg.norm(numer)
    U = numer / denom
    return U

def field_strength(k, band, Nk):
    """FHS lattice field strength"""
    T = [
        link1(k, band, Nk),
        link2(bump_x(k, Nk), band, Nk),
        1 / link1(bump_y(k, Nk), band, Nk ),
        1 / link2(k, band, Nk)
    ]
    F = np.log(np.prod(T))
    return  F

def calculate(band, Nk):
    """Compute the Chern number for a band using the FHS method"""
    F = [field_strength(k, band, Nk) for k in band]
    F = sum(F)
    F = np.real(F / (2 * np.pi * 1j))
    return F



# pylint: disable=line-too-long
"""
Miscellaneous utilities.
"""
import numpy as np

def ragged_to_ndarray(ragged_list, dtype=complex, value=0):
    # pylint: disable=raise-missing-from, consider-using-enumerate
    """
    Convert a list of ndarrays, of different shapes, into an ndarray
    via padding
    """

    # Find the largest subarray
    shapes = []
    for array in ragged_list:
        shapes.append(np.array(array).shape)
    try:
        shapes = np.array(shapes, dtype=int)
    except:
        raise ValueError("All subarrays must have the same number of indices")

    # Largest subarray is our target shape
    target_shape = np.zeros(len(shapes[0]), dtype=int)
    for i in range(len(shapes[0])):
        target_shape[i] = shapes[:, i].max()
    target_shape = tuple(target_shape)
    array = np.zeros(((len(ragged_list),) + target_shape), dtype=dtype)
    for i in range(len(ragged_list)):
        pad_width = [(0, ii - i) for i, ii in zip(np.array(ragged_list[i]).shape, target_shape)]
        array[i] = np.pad(ragged_list[i], pad_width=pad_width, mode="constant", constant_values=value)

    return array
